/*
 * rotate-sb-plugin.c -- Status bar plugin for randr control
 *
 * Copyright (C) 2008 Jonas Hurrelmann <j@outpo.st>
 *
 * Plugin parts where taken from vagalume-sb-plugin by
 * Copyright (C) 2008 Mario Sanchez Prada <msanchez@igalia.com>
 *
 * Randr handling was taken from gpe-conf
 * Copyright (C) 2002 Moray Allan <moray@sermisy.org>, Pierre TARDY <tardyp@free.fr>
 *               2003-2007 Florian Boor <florian.boor@kernelconcepts.de>
 *
 * This file is part of Rotate and is published under the GNU GPLv3
 * See the README file for more details.
 * Modified by qwerty12 to change d-pad maps when rotating. Thanks to Benson, pH5 & Johnx
 */

#include <string.h>
#include <gtk/gtk.h>
#include <libosso.h>
#include <hildon/hildon-banner.h>
#include <libhildonwm/hd-wm.h>
#include <libhildonwm/hd-wm-application.h>
#include <libhildondesktop/statusbar-item.h>
#include <libhildondesktop/libhildondesktop.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/extensions/Xrandr.h>
#include <stdlib.h>
#include "rotate-sb-plugin.h"

#define MAIN_PANEL_WIDTH 300

#define NORMAL_ITEM_STRING "Normal"
#define LEFT_ITEM_STRING "Left"
#define RIGHT_ITEM_STRING "Right"
#define INVERSED_ITEM_STRING "Inversed"

#define ICON_DIR "/usr/share/pixmaps"
#define APP_ICON ICON_DIR "/rotate.png"


#define ROTATE_SB_PLUGIN_GET_PRIVATE(object) \
                 (G_TYPE_INSTANCE_GET_PRIVATE ((object), \
                         ROTATE_SB_PLUGIN_TYPE, \
                         RotateSbPluginPrivate))

typedef struct _RotateSbPluginPrivate RotateSbPluginPrivate;
struct _RotateSbPluginPrivate
{
        osso_context_t *osso_context;

        GtkWidget *button;
        GtkWidget *icon;
        GtkWidget *main_panel;
        GtkWidget *normal_item;
        GtkWidget *left_item;
        GtkWidget *right_item;
        GtkWidget *inversed_item;

        gint plugin_btn_handler_id;
        gint main_panel_handler_id;
        gint normal_item_handler_id;
        gint left_item_handler_id;
        gint right_item_handler_id;
        gint inversed_item_handler_id;
};

HD_DEFINE_PLUGIN (RotateSbPlugin, rotate_sb_plugin, STATUSBAR_TYPE_ITEM);

/* Initialization/destruction functions */
static void rotate_sb_plugin_class_init (RotateSbPluginClass *klass);
static void rotate_sb_plugin_init (RotateSbPlugin *rsbp);
static void rotate_sb_plugin_finalize (GObject *object);

/* Panel update functions */
static void main_panel_create (RotateSbPlugin *rsbp);

/* Signals handlers */
static void plugin_btn_toggled (GtkWidget *button, gpointer data);
static void main_panel_item_activated (GtkWidget *item, gpointer data);
static void main_panel_hidden (GtkWidget *main_panel, gpointer user_data);

/* Randr Helper functions */
static gint get_rotation (void);
static void set_rotation (gint);
static gint check_init_rotation(void);

/* X Display and screen */
static Display *dpy = NULL;
static gint screen;

/* Initialization/destruction functions */

static void
rotate_sb_plugin_class_init (RotateSbPluginClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);
        object_class->finalize = rotate_sb_plugin_finalize;

        g_type_class_add_private (object_class,
                                  sizeof (RotateSbPluginPrivate));
}

static void
rotate_sb_plugin_init (RotateSbPlugin *rsbp)
{
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);

        priv->osso_context = NULL;

		/* Setup Randr */
		if (check_init_rotation()) {
				GdkPixbuf *icon_pixbuf;
				icon_pixbuf = gdk_pixbuf_new_from_file(APP_ICON, NULL);

				/* Init private attributes */
				priv->button = NULL;
				priv->icon = NULL;
			
				priv->normal_item = NULL;
				priv->left_item = NULL;
				priv->right_item = NULL;
				priv->inversed_item = NULL;
			
				/* Create status bar plugin button */
				priv->button = gtk_toggle_button_new ();
				priv->icon = gtk_image_new_from_pixbuf (icon_pixbuf);
				gtk_container_add (GTK_CONTAINER (priv->button), priv->icon);
				gtk_container_add (GTK_CONTAINER (rsbp), priv->button);

				/* Connect signals */
				priv->plugin_btn_handler_id =
						g_signal_connect(G_OBJECT(priv->button), "pressed",
										 G_CALLBACK(plugin_btn_toggled), rsbp);

				/* Show widgets */
				gtk_widget_show_all (priv->button);
				
				/* Create main panel */
				main_panel_create (rsbp);
		}
}

static void
rotate_sb_plugin_finalize (GObject *object)
{
        RotateSbPlugin *rsbp = ROTATE_SB_PLUGIN (object);
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);

        /* Destroy local widgets */
        if (priv->main_panel) {
                gtk_widget_destroy (priv->main_panel);
        }

        /* Disconnect handlers */
        g_signal_handler_disconnect (priv->button,
                                     priv->plugin_btn_handler_id);
        g_signal_handler_disconnect (priv->main_panel,
                                     priv->main_panel_handler_id);

		g_signal_handler_disconnect (priv->normal_item,
                                     priv->normal_item_handler_id);
        g_signal_handler_disconnect (priv->left_item,
                                     priv->left_item_handler_id);
        g_signal_handler_disconnect (priv->right_item,
                                     priv->right_item_handler_id);
        g_signal_handler_disconnect (priv->inversed_item,
                                     priv->inversed_item_handler_id);

        G_OBJECT_CLASS (g_type_class_peek_parent
                        (G_OBJECT_GET_CLASS(object)))->finalize(object);
}

static PangoFontDescription *
get_small_font (GtkWidget *widget)
{
        static PangoFontDescription *new_font = NULL;

        if (new_font == NULL) {
                new_font = pango_font_description_from_string
                        ("Nokia Sans 13.625");
        }

        return new_font;
}

static void
main_panel_create (RotateSbPlugin *rsbp)
{
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);
        GtkWidget *label;

        /* Create main_panel and main_panel items */
        priv->main_panel = gtk_menu_new ();
        /*gtk_widget_set_size_request (priv->main_panel, MAIN_PANEL_WIDTH, -1);*/

        GtkWidget *label_item = gtk_menu_item_new ();
	    label = gtk_label_new (NULL);
        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
        gtk_widget_modify_font (label, get_small_font (label));
        gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
        gtk_container_add (GTK_CONTAINER (label_item), label);
        gtk_widget_set_sensitive (label_item, FALSE);
	    gtk_label_set_markup (GTK_LABEL (label), "Rotation");
        label->state = GTK_STATE_NORMAL;

	
        /* Active items */
        priv->normal_item = gtk_menu_item_new_with_label (NORMAL_ITEM_STRING);
        priv->left_item = gtk_menu_item_new_with_label (LEFT_ITEM_STRING);
        priv->right_item = gtk_menu_item_new_with_label (RIGHT_ITEM_STRING);
        priv->inversed_item = gtk_menu_item_new_with_label (INVERSED_ITEM_STRING);

        /* Add items to main_panel */
	    gtk_menu_append (priv->main_panel, label_item);

        gtk_menu_append (priv->main_panel, priv->normal_item);
        gtk_menu_append (priv->main_panel, priv->left_item);
        gtk_menu_append (priv->main_panel, priv->right_item);
        gtk_menu_append (priv->main_panel, priv->inversed_item);

        /* Connect signals */
        priv->main_panel_handler_id =
                g_signal_connect(priv->main_panel,
                                 "selection-done",
                                 G_CALLBACK(main_panel_hidden), rsbp);
        priv->normal_item_handler_id =
                g_signal_connect(priv->normal_item, "activate",
                                 G_CALLBACK (main_panel_item_activated), rsbp);
        priv->left_item_handler_id =
                g_signal_connect(priv->left_item, "activate",
                                 G_CALLBACK (main_panel_item_activated), rsbp);
        priv->right_item_handler_id =
                g_signal_connect(priv->right_item, "activate",
                                 G_CALLBACK (main_panel_item_activated), rsbp);
        priv->inversed_item_handler_id =
                g_signal_connect(priv->inversed_item, "activate",
                                 G_CALLBACK (main_panel_item_activated), rsbp);
        /* Show widgets */
        gtk_widget_show_all (priv->main_panel);
}

/* Signals handlers */

static void
main_panel_position_func (GtkMenu   *main_panel,
                          gint      *x,
                          gint      *y,
                          gboolean  *push_in,
                          gpointer   data)
{
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (data);

        GtkRequisition req;

        gtk_widget_size_request (GTK_WIDGET (main_panel->toplevel), &req);

        gdk_window_get_origin (priv->button->window, x, y);
        *x += (priv->button->allocation.x + priv->button->allocation.width
               - req.width);
        *y += (priv->button->allocation.y + priv->button->allocation.height +
               10);
}

static void
plugin_btn_toggled (GtkWidget *button, gpointer data)
{
        RotateSbPlugin *rsbp = ROTATE_SB_PLUGIN (data);
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);

        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (button)))
                return;
	
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), True);
        g_debug ("Plugin button toggled");

        gtk_menu_popup (GTK_MENU (priv->main_panel),
                        NULL, NULL,
                        main_panel_position_func, rsbp,
                        1,
                        gtk_get_current_event_time ());
}

static void
main_panel_item_activated (GtkWidget *item, gpointer data)
{
        RotateSbPlugin *rsbp = ROTATE_SB_PLUGIN (data);
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);

        if (item == priv->normal_item) {
				set_rotation(0);
                g_debug ("Request sent: Normal");
        } else if (item == priv->left_item) {
				set_rotation(1);
                g_debug ("Request sent: Left");
        } else if (item == priv->right_item) {
				set_rotation(3);
                g_debug ("Request sent: Right");
        } else if (item == priv->inversed_item) {
				set_rotation(2);
                g_debug ("Request sent: Inversed");
        } else {
                g_debug ("Unknown action");
        }
}

static void
main_panel_hidden (GtkWidget *main_panel, gpointer data)
{
        RotateSbPlugin *rsbp = ROTATE_SB_PLUGIN (data);
        RotateSbPluginPrivate *priv = ROTATE_SB_PLUGIN_GET_PRIVATE (rsbp);

        g_debug ("Main panel hidden");

        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->button), FALSE);
}

/* Helper functions */
static gint
xrr_supported (void)
{
  int xrr_event_base, xrr_error_base;
  int xrr_major, xrr_minor;

  if (XRRQueryExtension (dpy, &xrr_event_base, &xrr_error_base) == False
      || XRRQueryVersion (dpy, &xrr_major, &xrr_minor) == 0
      || xrr_major != 1
      || xrr_minor < 1)
    return 0;

  return 1;
}


static gint
check_init_rotation(void)
{
	dpy = XOpenDisplay (NULL);
	if (dpy == NULL)
 	{
		fprintf (stderr, "Couldn't open display\n");
		return FALSE;
	}
	screen = DefaultScreen (dpy);
	
	return xrr_supported ();	
}
  

static gint 
get_rotation (void)
{
	gint rotation;
	XRRScreenConfiguration *rr_screen;
	Rotation current_rotation;
	
	if (dpy == NULL) 
		if (!check_init_rotation()) return 0;

	rr_screen = XRRGetScreenInfo (dpy, RootWindow (dpy, screen));
	XRRRotations (dpy, screen, &current_rotation);
	XRRFreeScreenConfigInfo (rr_screen);

	switch (current_rotation)
	{
	case RR_Rotate_270:
		rotation = 3;
	break;
	case RR_Rotate_180:
		rotation = 2;
	break;
	case RR_Rotate_90:
		rotation = 1;
	break;
	case RR_Rotate_0:
		rotation = 0;
	break;
	default:
		fprintf (stderr, "Unknown RANDR rotation: %d\n", current_rotation);
		rotation = 0;
	break;
	}
	
  return rotation;
}


static void 
set_rotation (gint rotation)
{
	Rotation sc_rotation;
	XRRScreenConfiguration *scr_config;
	Rotation current_rotation;
	gint size;

	if (dpy == NULL) 
		if (!check_init_rotation()) return;
	
	scr_config =  XRRGetScreenInfo (dpy,RootWindow (dpy, screen));
	size = XRRConfigCurrentConfiguration (scr_config, &current_rotation);	

	switch (rotation)
	{
	case 3:
		sc_rotation = RR_Rotate_270;
		system("xmodmap -e 'keycode 114 = Up' -e 'keycode 111 = Left' -e 'keycode 113 = Down' -e 'keycode 116 = Right'");
	break;
	case 2:
		sc_rotation = RR_Rotate_180;
		system("xmodmap -e 'keycode 114 = Left' -e 'keycode 111 = Down' -e 'keycode 113 = Right' -e 'keycode 116 = Up'");
	break;
	case 1:
		sc_rotation = RR_Rotate_90;
		system("xmodmap -e 'keycode 114 = Down' -e 'keycode 111 = Right' -e 'keycode 113 = Up' -e 'keycode 116 = Left'");
	break;
	case 0:
		sc_rotation = RR_Rotate_0;
		system("xmodmap -e 'keycode 114 = Right' -e 'keycode 111 = Up' -e 'keycode 113 = Left' -e 'keycode 116 = Down'");
	break;
	default:
		fprintf (stderr, "Unknown RANDR rotation: %d\n", rotation);
		sc_rotation = RR_Rotate_0;
	break;
	}
	XRRSetScreenConfig (dpy, 
			   scr_config,
			   RootWindow (dpy, screen),
			   size,
			   sc_rotation,
			   CurrentTime);
	
	XRRFreeScreenConfigInfo (scr_config);
}
